# -*- coding: utf-8 -*-
"""Mixin that adds support of recommended cutoffs to a ``Group`` subclass, using its extras."""
from typing import Union

from aiida.common.lang import type_check
from aiida.plugins import DataFactory

StructureData = DataFactory('structure')  # pylint: disable=invalid-name

__all__ = ('RecommendedCutoffMixin',)


class RecommendedCutoffMixin:
    """Mixin that adds support of recommended cutoffs to a ``Group`` subclass, using its extras.

    This class assumes that the cutoffs apply to a plane-wave based code and as such the cutoffs pertain to the wave
    functions and the charge density.
    """

    _key_cutoffs = '_cutoffs'
    _key_default_stringency = '_default_stringency'

    @classmethod
    def validate_cutoffs(cls, elements: set, cutoffs: dict) -> None:
        """Validate a cutoff dictionary for a given set of elements.

        :param elements: set of elements for which to validate the cutoffs dictionary.
        :param cutoffs: dictionary with recommended cutoffs. Format: multiple sets of recommended cutoffs can be
            specified where the key represents the stringency, e.g. ``low`` or ``normal``. For each stringency, a
            dictionary should be defined that for each element symbols for which the family contains a pseudopotential,
            two values are specified, ``cutoff_wfc`` and ``cutoff_rho``, containing a float value with the recommended
            cutoff to be used for the wave functions and charge density, respectively.
        :raises ValueError: if the set of elements and those defined in the cutoffs do not match exactly, or if the
            cutoffs dictionary has an invalid format.
        """
        elements_family = set(elements)

        for stringency, cutoffs_sub in cutoffs.items():

            elements_cutoffs = set(cutoffs_sub.keys())
            elements_diff = elements_family ^ elements_cutoffs

            if elements_family < elements_cutoffs:
                raise ValueError(f'cutoffs defined for unsupported elements: {elements_diff}')

            if elements_family > elements_cutoffs:
                raise ValueError(f'cutoffs not defined for all family elements: {elements_diff}')

            for element, values in cutoffs_sub.items():
                if set(values.keys()) != {'cutoff_wfc', 'cutoff_rho'}:
                    raise ValueError(
                        f'invalid cutoff keys for stringency `{stringency}` and element {element}: {values}'
                    )

                if any(not isinstance(cutoff, (int, float)) for cutoff in values.values()):
                    raise ValueError(
                        f'invalid cutoff values for stringency `{stringency}` and element {element}: {values}'
                    )

    def validate_stringency(self, stringency: str) -> None:
        """Validate a cutoff stringency.

        :param stringency: the cutoff stringency to validate.
        :raises ValueError: if stringency is None or the family does not define cutoffs for the specified stringency.
        """
        if stringency is None:
            raise ValueError('defining a stringency is required.')

        if stringency not in self.get_cutoff_stringencies():
            raise ValueError(f'stringency `{stringency}` is not defined for this family.')

    def _get_cutoffs(self) -> dict:
        """Return the cutoffs dictionary.

        :return: the cutoffs extra or an empty dictionary if it has not yet been set.
        """
        return self.get_extra(self._key_cutoffs, {})

    def get_default_stringency(self) -> Union[str, None]:
        """Return the default stringency if defined.

        :return: the default stringency or ``None`` if no cutoffs have been defined for this family.
        :raises ValueError: if not default stringency has been defined.
        """
        try:
            return self.get_extra(self._key_default_stringency)
        except AttributeError as exception:
            raise ValueError('no default stringency has been defined.') from exception

    def get_cutoff_stringencies(self) -> tuple:
        """Return a tuple of the available cutoff stringencies.

        :return: the tuple of stringencies that are defined for this family.
        """
        return tuple(self._get_cutoffs().keys())

    def set_cutoffs(self, cutoffs: dict, default_stringency: str = None) -> None:
        """Set the recommended cutoffs for the pseudos in this family.

        :param cutoffs: dictionary with recommended cutoffs. Format: one or multiple sets of recommended cutoffs can be
            specified where each key represents a stringency set, e.g. ``low`` or ``normal``. For each stringency, a
            dictionary should be defined that for each element symbols for which the family contains a pseudopotential,
            two values are specified, ``cutoff_wfc`` and ``cutoff_rho``, containing a float value with the recommended
            cutoff to be used for the wave functions and charge density, respectively.
        :param default_stringency: the default stringency to be used when ``get_recommended_cutoffs`` is called. If is
            possible to not specify this if and only if the cutoffs only contain a single stringency set. That one will
            then automatically be set as default.
        :raises ValueError: if the cutoffs have an invalid format or the default stringency is invalid.
        """
        self.validate_cutoffs(set(self.elements), cutoffs)

        if default_stringency is None and len(cutoffs) != 1:
            raise ValueError('have to explicitly specify a default stringency when specifying multiple cutoff sets.')

        default_stringency = default_stringency or list(cutoffs.keys())[0]

        self.set_extra(self._key_cutoffs, cutoffs)
        self.set_extra(self._key_default_stringency, default_stringency)

    def get_cutoffs(self, stringency=None) -> Union[dict, None]:
        """Return a set of cutoffs for the given stringency.

        :param stringency: optional stringency if different from the default.
        :return: the cutoffs or ``None`` if no cutoffs whatsoever have been defined for this family.
        :raises ValueError: if the requested stringency is not defined for this family.
        """
        stringency = stringency or self.get_default_stringency()

        try:
            return self._get_cutoffs()[stringency]
        except KeyError as exception:
            raise ValueError(f'stringency `{stringency}` is not defined for this family.') from exception

    def get_recommended_cutoffs(self, *, elements=None, structure=None, stringency=None):
        """Return tuple of recommended wavefunction and density cutoffs for the given elements or ``StructureData``.

        .. note:: at least one and only one of arguments ``elements`` or ``structure`` should be passed.

        :param elements: single or tuple of elements.
        :param structure: a ``StructureData`` node.
        :param stringency: optional stringency if different from the default.
        :return: tuple of recommended wavefunction and density cutoff.
        :raises ValueError: if the requested stringency is not defined for this family.
        """
        if (elements is None and structure is None) or (elements is not None and structure is not None):
            raise ValueError('at least one and only one of `elements` or `structure` should be defined')

        type_check(elements, (tuple, str), allow_none=True)
        type_check(structure, StructureData, allow_none=True)

        if structure is not None:
            symbols = structure.get_symbols_set()
        elif isinstance(elements, tuple):
            symbols = elements
        else:
            symbols = (elements,)

        cutoffs_wfc = []
        cutoffs_rho = []
        cutoffs = self.get_cutoffs(stringency=stringency)

        for element in symbols:
            values = cutoffs[element]
            cutoffs_wfc.append(values['cutoff_wfc'])
            cutoffs_rho.append(values['cutoff_rho'])

        return (max(cutoffs_wfc), max(cutoffs_rho))



# -*- coding: utf-8 -*-
"""Group to represent a pseudo potential family with pseudos in PSML format."""
from aiida.plugins import DataFactory

from .pseudo import PseudoPotentialFamily

__all__ = ('PsmlFamily',)

PsmlData = DataFactory('pseudo.psml')  # pylint: disable=invalid-name


class PsmlFamily(PseudoPotentialFamily):
    """Group to represent a pseudo potential family with pseudos in PSML format."""

    _pseudo_type = PsmlData



# -*- coding: utf-8 -*-
"""Tests for the :mod:`~aiida_pseudo.cli.params.types` module."""
import click
import pytest

from aiida_pseudo.cli.params import types
from aiida_pseudo.groups.family import PseudoPotentialFamily


def test_pseudo_family_type_param_convert(ctx):
    """Test the `PseudoPotentialFamilyTypeParam.convert` method."""
    param = types.PseudoPotentialFamilyTypeParam()

    with pytest.raises(click.BadParameter, match=r'.*is not an existing group plugin.'):
        param.convert('non.existing', None, ctx)

    with pytest.raises(click.BadParameter, match=r'.*entry point is not a subclass of `PseudoPotentialFamily`.'):
        param.convert('core', None, ctx)

    assert param.convert('pseudo.family', None, ctx) is PseudoPotentialFamily


def test_pseudo_family_type_param_complete(ctx):
    """Test the `PseudoPotentialFamilyTypeParam.complete` method."""
    param = types.PseudoPotentialFamilyTypeParam()
    assert isinstance(param.complete(ctx, ''), list)
    assert isinstance(param.complete(ctx, 'pseudo'), list)
    assert ('pseudo.family', '') in param.complete(ctx, 'pseudo')



# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name
"""Tests for the :mod:`aiida_pseudo.groups.mixins.cutoffs` module."""
import copy

import pytest

from aiida.orm.groups import GroupMeta
from aiida_pseudo.groups.family import PseudoPotentialFamily
from aiida_pseudo.groups.mixins import RecommendedCutoffMixin


class FixtureGroupMeta(GroupMeta):
    """Meta class for `aiida.orm.groups.Group` to automatically set the `type_string` attribute."""

    def __new__(mcs, name, bases, namespace, **kwargs):  # pylint: disable=bad-mcs-classmethod-argument
        """Construct new instance of the class."""
        newcls = GroupMeta.__new__(mcs, name, bases, namespace, **kwargs)  # pylint: disable=too-many-function-args
        newcls._type_string = 'test'  # pylint: disable=protected-access
        return newcls


class CutoffFamily(RecommendedCutoffMixin, PseudoPotentialFamily, metaclass=FixtureGroupMeta):
    """Test class for the ``RecommendedCutoffMixin``."""


@pytest.fixture
def get_cutoffs():
    """Return a cutoffs dictionary."""

    def _get_cutoffs(family, stringencies=('default',)):
        cutoffs = {}

        for element in family.elements:
            cutoffs[element] = {
                'cutoff_wfc': 1.0,
                'cutoff_rho': 2.0,
            }

        return {stringency: cutoffs for stringency in stringencies}

    return _get_cutoffs


@pytest.mark.usefixtures('clear_db')
def test_get_cutoffs_private(get_pseudo_family, get_cutoffs):
    """Test the ``RecommendedCutoffMixin._get_cutoffs`` method."""
    family = get_pseudo_family(cls=CutoffFamily)
    assert family._get_cutoffs() == {}  # pylint: disable=protected-access

    family.set_cutoffs(get_cutoffs(family), 'default')
    assert family._get_cutoffs() == get_cutoffs(family)  # pylint: disable=protected-access


@pytest.mark.usefixtures('clear_db')
def test_validate_stringency(get_pseudo_family, get_cutoffs):
    """Test the ``RecommendedCutoffMixin.validate_stringency`` method."""
    family = get_pseudo_family(cls=CutoffFamily)

    with pytest.raises(ValueError, match=r'stringency `.*` is not defined for this family.'):
        family.validate_stringency('default')

    cutoffs = get_cutoffs(family)
    stringency = list(cutoffs.keys())[0]
    family.set_cutoffs(cutoffs, stringency)

    with pytest.raises(ValueError, match=r'stringency `.*` is not defined for this family.'):
        family.validate_stringency(stringency + 'non-existing')

    family.validate_stringency(stringency)


@pytest.mark.usefixtures('clear_db')
def test_get_default_stringency(get_pseudo_family, get_cutoffs):
    """Test the ``RecommendedCutoffMixin.get_default_stringency`` method."""
    family = get_pseudo_family(cls=CutoffFamily)

    with pytest.raises(ValueError, match='no default stringency has been defined.'):
        family.get_default_stringency()

    cutoffs = get_cutoffs(family)
    stringency = list(cutoffs.keys())[0]
    family.set_cutoffs(cutoffs, stringency)

    assert family.get_default_stringency() == stringency


@pytest.mark.usefixtures('clear_db')
def test_get_cutoff_stringencies(get_pseudo_family, get_cutoffs):
    """Test the ``RecommendedCutoffMixin.get_cutoff_stringencies`` method."""
    family = get_pseudo_family(cls=CutoffFamily)
    assert family.get_cutoff_stringencies() == ()

    stringencies = ('low', 'normal', 'high')
    cutoffs = get_cutoffs(family, stringencies)
    family.set_cutoffs(cutoffs, stringencies[0])

    assert sorted(family.get_cutoff_stringencies()) == sorted(stringencies)


@pytest.mark.usefixtures('clear_db')
def test_set_cutoffs(get_pseudo_family):
    """Test the `RecommendedCutoffMixin.set_cutoffs` method."""
    elements = ['Ar', 'He']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffFamily, elements=elements)
    cutoffs = {'default': {element: {'cutoff_wfc': 1.0, 'cutoff_rho': 2.0} for element in elements}}
    family.set_cutoffs(cutoffs, 'default')

    assert family.get_cutoffs() == cutoffs['default']

    with pytest.raises(ValueError, match=r'cutoffs defined for unsupported elements: .*'):
        cutoffs_invalid = copy.deepcopy(cutoffs)
        cutoffs_invalid['default']['C'] = {'cutoff_wfc': 1.0, 'cutoff_rho': 2.0}
        family.set_cutoffs(cutoffs_invalid, 'default')

    with pytest.raises(ValueError, match=r'cutoffs not defined for all family elements: .*'):
        cutoffs_invalid = copy.deepcopy(cutoffs)
        cutoffs_invalid['default'].pop('He')
        family.set_cutoffs(cutoffs_invalid, 'default')

    with pytest.raises(ValueError, match=r'invalid cutoff keys for stringency `default` and element .*: .*'):
        cutoffs_invalid = copy.deepcopy(cutoffs)
        cutoffs_invalid['default']['He'] = {'cutoff_wfc': 1.0}
        family.set_cutoffs(cutoffs_invalid, 'default')

    with pytest.raises(ValueError, match=r'invalid cutoff keys for stringency `default` and element .*: .*'):
        cutoffs_invalid = copy.deepcopy(cutoffs)
        cutoffs_invalid['default']['He'] = {'cutoff_wfc': 1.0, 'cutoff_rho': 2.0, 'cutoff_extra': 3.0}
        family.set_cutoffs(cutoffs_invalid, 'default')

    with pytest.raises(ValueError, match=r'invalid cutoff values for stringency `default` and element .*: .*'):
        cutoffs_invalid = copy.deepcopy(cutoffs)
        cutoffs_invalid['default']['He'] = {'cutoff_wfc': 1.0, 'cutoff_rho': 'string'}
        family.set_cutoffs(cutoffs_invalid, 'default')


@pytest.mark.usefixtures('clear_db')
def test_set_cutoffs_auto_default(get_pseudo_family):
    """Test the `RecommendedCutoffMixin.set_cutoffs` method when not specifying explicit default.

    If the cutoffs specified only contain a single set, the `default_stringency` is determined automatically.
    """
    elements = ['Ar']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffFamily, elements=elements)
    values = {element: {'cutoff_wfc': 1.0, 'cutoff_rho': 2.0} for element in elements}
    cutoffs = {'default': values}

    with pytest.raises(ValueError, match='specify a default stringency when specifying multiple cutoff sets.'):
        family.set_cutoffs({'low': values, 'hi': values})

    family.set_cutoffs(cutoffs)
    assert family.get_cutoffs() == cutoffs['default']
    assert family.get_default_stringency() == 'default'


@pytest.mark.usefixtures('clear_db')
def test_get_cutoffs(get_pseudo_family):
    """Test the `RecommendedCutoffMixin.get_cutoffs` method."""
    elements = ['Ar', 'He']
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffFamily, elements=elements)
    cutoffs = {'default': {element: {'cutoff_wfc': 1.0, 'cutoff_rho': 2.0} for element in elements}}

    with pytest.raises(ValueError, match='no default stringency has been defined.'):
        family.get_cutoffs()

    family.set_cutoffs(cutoffs, 'default')

    with pytest.raises(ValueError, match=r'stringency `.*` is not defined for this family.'):
        family.get_cutoffs('non-existing')

    assert family.get_cutoffs() == cutoffs['default']


@pytest.mark.usefixtures('clear_db')
def test_get_recommended_cutoffs(get_pseudo_family, generate_structure):
    """Test the `RecommendedCutoffMixin.get_recommended_cutoffs` method."""
    elements = ['Ar', 'He']
    cutoffs = {
        'default': {
            'Ar': {
                'cutoff_wfc': 1.0,
                'cutoff_rho': 2.0
            },
            'He': {
                'cutoff_wfc': 3.0,
                'cutoff_rho': 8.0
            },
        }
    }
    family = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=CutoffFamily, elements=elements)
    family.set_cutoffs(cutoffs, 'default')
    structure = generate_structure(elements=elements)

    with pytest.raises(ValueError):
        family.get_recommended_cutoffs(elements=None, structure=None)

    with pytest.raises(ValueError):
        family.get_recommended_cutoffs(elements='Ar', structure=structure)

    with pytest.raises(TypeError):
        family.get_recommended_cutoffs(elements=False, structure=None)

    with pytest.raises(TypeError):
        family.get_recommended_cutoffs(elements=None, structure='Ar')

    expected = cutoffs['default']['Ar']
    assert family.get_recommended_cutoffs(elements='Ar') == (expected['cutoff_wfc'], expected['cutoff_rho'])
    assert family.get_recommended_cutoffs(elements=('Ar',)) == (expected['cutoff_wfc'], expected['cutoff_rho'])

    expected = cutoffs['default']['He']
    assert family.get_recommended_cutoffs(elements=('Ar', 'He')) == (expected['cutoff_wfc'], expected['cutoff_rho'])

    expected = cutoffs['default']['He']
    assert family.get_recommended_cutoffs(structure=structure) == (expected['cutoff_wfc'], expected['cutoff_rho'])

    # Try structure with multiple kinds with the same element
    expected = cutoffs['default']['He']
    structure = generate_structure(elements=['He1', 'He2'])
    assert family.get_recommended_cutoffs(structure=structure) == (expected['cutoff_wfc'], expected['cutoff_rho'])



# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name,unused-argument
"""Configuration and fixtures for unit test suite."""
import io
import os
import re
import shutil

import click
import pytest

from aiida.plugins import DataFactory

from aiida_pseudo.data.pseudo import PseudoPotentialData
from aiida_pseudo.groups.family import PseudoPotentialFamily

pytest_plugins = ['aiida.manage.tests.pytest_fixtures']  # pylint: disable=invalid-name


@pytest.fixture
def clear_db(clear_database_before_test):
    """Alias for the `clear_database_before_test` fixture from `aiida-core`."""
    yield


@pytest.fixture
def ctx():
    """Return an empty `click.Context` instance."""
    return click.Context(click.Command(name='dummy'))


@pytest.fixture
def run_cli_command():
    """Run a `click` command with the given options.

    The call will raise if the command triggered an exception or the exit code returned is non-zero.
    """

    def _run_cli_command(command, options=None, raises=None):
        """Run the command and check the result.

        :param command: the command to invoke
        :param options: the list of command line options to pass to the command invocation
        :param raises: optionally an exception class that is expected to be raised
        """
        import traceback
        from click.testing import CliRunner

        runner = CliRunner()
        result = runner.invoke(command, options or [])

        if raises is not None:
            assert result.exception is not None, result.output
            assert result.exit_code != 0
        else:
            assert result.exception is None, ''.join(traceback.format_exception(*result.exc_info))
            assert result.exit_code == 0, result.output

        result.output_lines = [line.strip() for line in result.output.split('\n') if line.strip()]

        return result

    return _run_cli_command


@pytest.fixture
def filepath_fixtures() -> str:
    """Return the absolute filepath to the directory containing the file `fixtures`.

    :return: absolute filepath to directory containing test fixture data.
    """
    return os.path.join(os.path.dirname(__file__), 'fixtures')


@pytest.fixture
def filepath_pseudos(filepath_fixtures):
    """Return the absolute filepath to the directory containing the pseudo potential files.

    :return: absolute filepath to directory containing test pseudo potentials.
    """

    def _filepath_pseudos(entry_point='upf') -> str:
        """Return the absolute filepath containing the pseudo potential files for a given entry point.

        :param entry_point: pseudo potential data entry point
        :return: filepath to folder containing pseudo files.
        """
        return os.path.join(filepath_fixtures, 'pseudos', entry_point)

    return _filepath_pseudos


@pytest.fixture
def get_pseudo_potential_data(filepath_pseudos):
    """Return a factory for `PseudoPotentialData` nodes."""

    def _get_pseudo_potential_data(element='Ar', entry_point=None) -> PseudoPotentialData:
        """Return a `PseudoPotentialData` for the given element.

        :param element: one of the elements for which there is a UPF test file available.
        :return: the `PseudoPotentialData`
        """
        if entry_point is None:
            cls = DataFactory('pseudo')
            pseudo = cls(io.BytesIO(b'content'), f'{element}.pseudo')
            pseudo.element = element
        else:
            cls = DataFactory(f'pseudo.{entry_point}')
            filename = f'{element}.{entry_point}'
            with open(os.path.join(filepath_pseudos(entry_point), filename), 'rb') as handle:
                pseudo = cls(handle, filename)

        return pseudo

    return _get_pseudo_potential_data


@pytest.fixture
def get_pseudo_family(tmpdir, filepath_pseudos):
    """Return a factory for a `PseudoPotentialFamily` instance."""

    def _get_pseudo_family(label='family', cls=PseudoPotentialFamily, elements=None) -> PseudoPotentialFamily:
        """Return an instance of `PseudoPotentialFamily` or subclass containing the given elements.

        :param elements: optional list of elements to include instead of all the available ones
        :return: the pseudo family
        """
        if elements is not None:
            elements = {re.sub('[0-9]+', '', element) for element in elements}

        dirpath = filepath_pseudos('upf')
        for pseudo in os.listdir(dirpath):
            if elements is None or any([pseudo.startswith(element) for element in elements]):
                shutil.copyfile(os.path.join(dirpath, pseudo), os.path.join(str(tmpdir), pseudo))

        return cls.create_from_folder(str(tmpdir), label)

    return _get_pseudo_family


@pytest.fixture
def get_pseudo_archive(tmpdir, filepath_pseudos):
    """Create an archive with pseudos."""

    def _get_pseudo_archive(fmt='gztar'):
        shutil.make_archive(str(tmpdir / 'archive'), fmt, filepath_pseudos('upf'))
        filepath = os.path.join(str(tmpdir), os.listdir(str(tmpdir))[0])
        yield filepath

    return _get_pseudo_archive


@pytest.fixture
def generate_structure():
    """Return a ``StructureData``."""

    def _generate_structure(elements=('Ar',)):
        """Return a ``StructureData``."""
        from aiida.orm import StructureData

        structure = StructureData(cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]])

        for index, element in enumerate(elements):
            symbol = re.sub(r'[0-9]+', '', element)
            structure.append_atom(position=(index * 0.5, index * 0.5, index * 0.5), symbols=symbol, name=element)

        return structure

    return _generate_structure



# -*- coding: utf-8 -*-
"""Test the root command of the CLI."""
from aiida_pseudo.cli import cmd_root


def test_root(run_cli_command):
    """Test the root command for the CLI is callable."""
    run_cli_command(cmd_root)

    for option in ['-h', '--help']:
        result = run_cli_command(cmd_root, [option])
        assert cmd_root.__doc__ in result.output



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""Test the command line interface utilities."""
import shutil
import tarfile
import tempfile

import pytest

from aiida_pseudo.cli.utils import attempt, create_family_from_archive
from aiida_pseudo.groups.family import PseudoPotentialFamily


@pytest.mark.usefixtures('clear_db')
@pytest.mark.parametrize(('fmt',), [(fmt[0],) for fmt in shutil.get_archive_formats()])
def test_create_family_from_archive(get_pseudo_archive, fmt):
    """Test the `create_family_from_archive` utility function."""
    label = 'PSEUDO/0.0/LDA/extreme'
    filepath_archive = next(get_pseudo_archive(fmt))
    family = create_family_from_archive(PseudoPotentialFamily, label, filepath_archive, fmt=fmt)

    assert isinstance(family, PseudoPotentialFamily)
    assert family.label == label
    assert family.count() != 0


@pytest.mark.usefixtures('clear_db')
def test_create_family_from_archive_incorrect_filetype(tmpdir):
    """Test the `create_family_from_archive` utility function for incorrect archive filetype."""
    with pytest.raises(OSError, match=r'failed to unpack the archive.*'):
        create_family_from_archive(PseudoPotentialFamily, 'label', str(tmpdir))


@pytest.mark.usefixtures('clear_db')
def test_create_family_from_archive_incorrect_format(tmpdir):
    """Test the `create_family_from_archive` utility function for invalid archive content."""
    with tempfile.NamedTemporaryFile(suffix='.tar.gz') as filepath_archive:

        with tarfile.open(filepath_archive.name, 'w:gz') as tar:
            tar.add(str(tmpdir), arcname='.')

        with pytest.raises(OSError, match=r'failed to parse pseudos from.*'):
            create_family_from_archive(PseudoPotentialFamily, 'label', filepath_archive.name)


def test_attempt_sucess(capsys):
    """Test the `attempt` utility function."""
    message = 'some message'

    with attempt(message):
        pass

    captured = capsys.readouterr()
    assert captured.out == f'Info: {message} [OK]\n'
    assert captured.err == ''


def test_attempt_exception(capsys):
    """Test the `attempt` utility function when exception is raised."""
    message = 'some message'
    exception = 'run-time-error'

    with pytest.raises(SystemExit):
        with attempt(message):
            raise RuntimeError(exception)

    captured = capsys.readouterr()
    assert captured.out == f'Info: {message} [FAILED]\n'
    assert captured.err == f'Critical: {exception}\n'


def test_attempt_exception_traceback(capsys):
    """Test the `attempt` utility function when exception is raised and `include_traceback=True`."""
    message = 'some message'
    exception = 'run-time-error'

    with pytest.raises(SystemExit):
        with attempt(message, include_traceback=True):
            raise RuntimeError(exception)

    captured = capsys.readouterr()
    assert captured.out == f'Info: {message} [FAILED]\n'
    assert captured.err.startswith(f'Critical: {exception}\n')
    assert 'Traceback' in captured.err



# -*- coding: utf-8 -*-
"""Tests for the :py:mod:`~aiida_pseudo.data.pseudo.pseudo` module."""
import io

import pytest

from aiida.common.files import md5_from_filelike
from aiida.common.exceptions import ModificationNotAllowed, StoringNotAllowed
from aiida.common.links import LinkType
from aiida.orm import CalcJobNode

from aiida_pseudo.data.pseudo import PseudoPotentialData


@pytest.mark.usefixtures('clear_db')
def test_constructor():
    """Test the constructor."""
    stream = io.BytesIO(b'pseudo')

    pseudo = PseudoPotentialData(stream)
    assert isinstance(pseudo, PseudoPotentialData)
    assert not pseudo.is_stored


@pytest.mark.usefixtures('clear_db')
def test_constructor_invalid():
    """Test the constructor for invalid arguments."""
    with pytest.raises(TypeError, match='missing 1 required positional argument'):
        PseudoPotentialData()  # pylint: disable=no-value-for-parameter


@pytest.mark.usefixtures('clear_db')
def test_store():
    """Test the `PseudoPotentialData.store` method."""
    stream = io.BytesIO(b'pseudo')
    md5_correct = md5_from_filelike(stream)
    md5_incorrect = 'abcdef0123456789'
    stream.seek(0)

    pseudo = PseudoPotentialData(io.BytesIO(b'pseudo'))

    with pytest.raises(StoringNotAllowed, match='no valid element has been defined.'):
        pseudo.store()

    pseudo.element = 'Ar'
    pseudo.set_attribute(PseudoPotentialData._key_md5, md5_incorrect)  # pylint: disable=protected-access

    with pytest.raises(StoringNotAllowed, match=r'md5 does not match that of stored file:'):
        pseudo.store()

    pseudo.md5 = md5_correct
    result = pseudo.store()
    assert result is pseudo
    assert pseudo.is_stored


@pytest.mark.usefixtures('clear_db')
def test_element():
    """Test the `PseudoPotentialData.element` property."""
    element = 'Ar'
    pseudo = PseudoPotentialData(io.BytesIO(b'pseudo'))
    assert pseudo.element is None

    element = 'He'
    pseudo.element = element
    assert pseudo.element == element

    with pytest.raises(ValueError, match=r'.* is not a valid element'):
        pseudo.element = 'Aa'

    pseudo.store()

    with pytest.raises(ModificationNotAllowed, match='the attributes of a stored entity are immutable'):
        pseudo.element = element


@pytest.mark.usefixtures('clear_db')
def test_md5():
    """Test the `PseudoPotentialData.md5` property."""
    stream = io.BytesIO(b'pseudo')
    md5 = md5_from_filelike(stream)
    stream.seek(0)

    pseudo = PseudoPotentialData(stream)
    pseudo.element = 'Ar'
    assert pseudo.md5 == md5

    with pytest.raises(ValueError, match=r'md5 does not match that of stored file.*'):
        pseudo.md5 = 'abcdef0123456789'

    pseudo.store()

    with pytest.raises(ModificationNotAllowed, match='the attributes of a stored entity are immutable'):
        pseudo.md5 = md5


@pytest.mark.usefixtures('clear_db')
def test_store_indirect():
    """Test the `PseudoPotentialData.store` method when called indirectly because its is an input."""
    pseudo = PseudoPotentialData(io.BytesIO(b'pseudo'))
    pseudo.element = 'Ar'

    node = CalcJobNode()
    node.add_incoming(pseudo, link_type=LinkType.INPUT_CALC, link_label='pseudo')
    node.store_all()



# -*- coding: utf-8 -*-
"""Command to install a pseudo potential family."""
import click

from aiida.cmdline.params import options as options_core
from aiida.cmdline.utils import decorators, echo

from .params import PseudoPotentialFamilyParam
from .root import cmd_root


@cmd_root.command('show')
@click.argument('family', type=PseudoPotentialFamilyParam(sub_classes=('aiida.groups:pseudo.family',)))
@options_core.RAW()
@decorators.with_dbenv()
def cmd_show(family, raw):
    """Show details of pseudo potential family."""
    from tabulate import tabulate

    rows = [[pseudo.element, pseudo.filename, pseudo.md5] for pseudo in family.nodes]
    headers = ['Element', 'Pseudo', 'MD5']

    if raw:
        echo.echo(tabulate(sorted(rows), tablefmt='plain'))
    else:
        echo.echo(tabulate(sorted(rows), headers=headers))



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""Tests for the command `aiida-pseudo list`."""
from aiida_pseudo.cli import cmd_list
from aiida_pseudo.cli.list import PROJECTIONS_VALID
from aiida_pseudo.groups.family import PseudoPotentialFamily, SsspFamily, UpfFamily


def test_list(clear_db, run_cli_command, get_pseudo_family):
    """Test the `aiida-pseudo list` command."""
    result = run_cli_command(cmd_list)
    assert 'no pseudo potential families have been installed yet: use `aiida-pseudo install`.' in result.output

    label = 'SSSP/1.0/PBE/efficiency'
    family = get_pseudo_family(label=label)
    result = run_cli_command(cmd_list)

    assert family.label in result.output


def test_list_raw(clear_db, run_cli_command, get_pseudo_family):
    """Test the `-r/--raw` option."""
    get_pseudo_family()

    for option in ['-r', '--raw']:
        result = run_cli_command(cmd_list, [option])
        assert len(result.output_lines) == 1


def test_list_project(clear_db, run_cli_command, get_pseudo_family):
    """Test the `-p/--project` option."""
    family = get_pseudo_family()

    # Test that all `PROJECTIONS_VALID` can actually be projected and don't except
    for option in ['-P', '--project']:
        run_cli_command(cmd_list, [option] + list(PROJECTIONS_VALID))

    result = run_cli_command(cmd_list, ['--raw', '-P', 'label'])
    assert len(result.output_lines) == 1
    assert family.label in result.output


def test_list_filter(clear_db, run_cli_command, get_pseudo_family):
    """Test the filtering options `--version`, `--functional` and `--protocol`."""
    family_upf = get_pseudo_family(label='UPF', cls=UpfFamily)
    family_sssp = get_pseudo_family(label='SSSP/1.0/PBE/efficiency', cls=SsspFamily)

    assert PseudoPotentialFamily.objects.count() == 2

    result = run_cli_command(cmd_list, ['--raw'])
    assert len(result.output_lines) == 2
    assert family_upf.label in result.output
    assert family_sssp.label in result.output

    result = run_cli_command(cmd_list, ['--raw', '-T', 'pseudo.family.sssp'])
    assert len(result.output_lines) == 1
    assert family_upf.label not in result.output
    assert family_sssp.label in result.output

    result = run_cli_command(cmd_list, ['--raw', '-T', 'pseudo.family'])
    assert 'no pseudo potential families found that match the filtering criteria.' in result.output



# -*- coding: utf-8 -*-
"""Group to represent a pseudo potential family with pseudos in UPF format."""
from aiida.plugins import DataFactory

from ..mixins import RecommendedCutoffMixin
from .pseudo import PseudoPotentialFamily

__all__ = ('UpfFamily',)

UpfData = DataFactory('pseudo.upf')  # pylint: disable=invalid-name


class UpfFamily(RecommendedCutoffMixin, PseudoPotentialFamily):
    """Group to represent a pseudo potential family with pseudos in UPF format."""

    _pseudo_type = UpfData



# -*- coding: utf-8 -*-
"""Base class for data types representing pseudo potentials."""
import typing

from aiida.common.constants import elements
from aiida.common.exceptions import StoringNotAllowed
from aiida.common.files import md5_from_filelike
from aiida.plugins import DataFactory

SingleFileData = DataFactory('singlefile')  # pylint: disable=invalid-name

__all__ = ('PseudoPotentialData',)


class PseudoPotentialData(SingleFileData):
    """Base class for data types representing pseudo potentials."""

    _key_element = 'element'
    _key_md5 = 'md5'

    @classmethod
    def validate_element(cls, element: str):
        """Validate the given element symbol.

        :param element: the symbol of the element following the IUPAC naming standard.
        :raises ValueError: if the element symbol is invalid.
        """
        if element not in [values['symbol'] for values in elements.values()]:
            raise ValueError(f'`{element}` is not a valid element.')

    def validate_md5(self, md5: str):
        """Validate that the md5 checksum matches that of the currently stored file.

        :param value: the md5 checksum.
        :raises ValueError: if the md5 does not match that of the currently stored file.
        """
        with self.open(mode='rb') as handle:
            md5_file = md5_from_filelike(handle)
            if md5 != md5_file:
                raise ValueError(f'md5 does not match that of stored file: {md5} != {md5_file}')

    def set_file(self, stream: typing.BinaryIO, filename: str = None, **kwargs):
        """Set the file content.

        :param stream: a filelike object with the binary content of the file.
        :param filename: optional explicit filename to give to the file stored in the repository.
        """
        super().set_file(stream, filename, **kwargs)
        stream.seek(0)
        self.md5 = md5_from_filelike(stream)

    def store(self, **kwargs):
        """Store the node verifying first that all required attributes are set.

        :raises :py:exc:`~aiida.common.StoringNotAllowed`: if no valid element has been defined.
        """
        try:
            self.validate_element(self.element)
        except ValueError as exception:
            raise StoringNotAllowed('no valid element has been defined.') from exception

        try:
            self.validate_md5(self.md5)
        except ValueError as exception:
            raise StoringNotAllowed(exception) from exception

        return super().store(**kwargs)

    @property
    def element(self) -> typing.Union[str, None]:
        """Return the element symbol.

        :return: the symbol of the element following the IUPAC naming standard or None if not defined.
        """
        return self.get_attribute(self._key_element, None)

    @element.setter
    def element(self, value: str):
        """Set the element.

        :param value: the symbol of the element following the IUPAC naming standard.
        :raises ValueError: if the element symbol is invalid.
        """
        self.validate_element(value)
        self.set_attribute(self._key_element, value)

    @property
    def md5(self) -> typing.Union[str, None]:
        """Return the md5.

        :return: the md5 of the stored file.
        """
        return self.get_attribute(self._key_md5, None)

    @md5.setter
    def md5(self, value: str):
        """Set the md5.

        :param value: the md5 checksum.
        :raises ValueError: if the md5 does not match that of the currently stored file.
        """
        self.validate_md5(value)
        self.set_attribute(self._key_md5, value)



# -*- coding: utf-8 -*-
"""Group to represent a pseudo potential family with pseudos in PSF format."""
from aiida.plugins import DataFactory

from .pseudo import PseudoPotentialFamily

__all__ = ('PsfFamily',)

PsfData = DataFactory('pseudo.psf')  # pylint: disable=invalid-name


class PsfFamily(PseudoPotentialFamily):
    """Group to represent a pseudo potential family with pseudos in PSF format."""

    _pseudo_type = PsfData



# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name
"""Tests for the `PseudoPotentialFamily` class."""
import distutils.dir_util
import os

import pytest

from aiida.common import exceptions
from aiida.orm import QueryBuilder

from aiida_pseudo.data.pseudo import PseudoPotentialData
from aiida_pseudo.groups.family.pseudo import PseudoPotentialFamily


def test_type_string():
    """Verify the `_type_string` class attribute is correctly set to the corresponding entry point name."""
    assert PseudoPotentialFamily._type_string == 'pseudo.family'  # pylint: disable=protected-access


def test_pseudo_type():
    """Test the `PseudoPotentialFamily.pseudo_type` method."""
    assert PseudoPotentialFamily.pseudo_type is PseudoPotentialData


@pytest.mark.filterwarnings('ignore:no registered entry point for')
def test_pseudo_type_validation():
    """Test that the constructor raises if `_pseudo_type` is not a subclass of `PseudoPotentialData`."""

    class CustomFamily(PseudoPotentialFamily):
        """Test subclass that intentionally defines incorrect type for `_pseudo_type`."""

        _pseudo_type = int

    with pytest.raises(RuntimeError, match=r'`.*` is not a subclass of `PseudoPotentialData`.'):
        CustomFamily(label='custom')


@pytest.mark.usefixtures('clear_db')
def test_construct():
    """Test the construction of `PseudoPotentialFamily` works."""
    label = 'label'
    family = PseudoPotentialFamily(label=label)
    assert isinstance(family, PseudoPotentialFamily)
    assert not family.is_stored
    assert family.label == label

    label = 'family'
    description = 'description'
    family = PseudoPotentialFamily(label=label, description=description)
    assert isinstance(family, PseudoPotentialFamily)
    assert not family.is_stored
    assert family.label == label
    assert family.description == description


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder(filepath_pseudos):
    """Test the `PseudoPotentialFamily.create_from_folder` class method."""
    label = 'label'
    family = PseudoPotentialFamily.create_from_folder(filepath_pseudos(), label)
    assert isinstance(family, PseudoPotentialFamily)
    assert family.is_stored
    assert family.label == label
    assert len(family.nodes) == len(os.listdir(filepath_pseudos()))


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_nested(filepath_pseudos, tmpdir):
    """Test the `PseudoPotentialFamily.create_from_folder` class method when the pseudos are in a subfolder."""
    filepath = str(tmpdir / 'subdirectory')
    distutils.dir_util.copy_tree(filepath_pseudos(), filepath)

    label = 'label'
    family = PseudoPotentialFamily.create_from_folder(str(tmpdir), label)
    assert isinstance(family, PseudoPotentialFamily)
    assert family.is_stored
    assert family.label == label
    assert len(family.nodes) == len(os.listdir(filepath_pseudos()))


@pytest.mark.usefixtures('clear_db')
@pytest.mark.parametrize('deduplicate', (True, False))
def test_create_from_folder_deduplicate(filepath_pseudos, deduplicate):
    """Test the `PseudoPotentialFamily.create_from_folder` class method."""
    from aiida_pseudo.groups.family.upf import UpfFamily

    # We create an existing `PseudoPotentialFamily` as well as a `UpfFamily` to test that the deduplication mechanism
    # will only ever check for pseudo potentials of the exact same type and not allow subclasses
    original = PseudoPotentialFamily.create_from_folder(filepath_pseudos(), 'original_family')
    UpfFamily.create_from_folder(filepath_pseudos(), 'upf_family')

    family = PseudoPotentialFamily.create_from_folder(filepath_pseudos(), 'duplicate_family', deduplicate=deduplicate)

    assert isinstance(family, PseudoPotentialFamily)
    assert family.is_stored

    pseudo_count = len(os.listdir(filepath_pseudos()))
    original_pseudos = {pseudo.pk for pseudo in original.pseudos.values()}
    family_pseudos = {pseudo.pk for pseudo in family.pseudos.values()}

    if deduplicate:
        assert QueryBuilder().append(PseudoPotentialFamily.pseudo_type, subclassing=False).count() == pseudo_count
        assert not original_pseudos.difference(family_pseudos)
    else:
        assert QueryBuilder().append(PseudoPotentialFamily.pseudo_type, subclassing=False).count() == pseudo_count * 2
        assert not original_pseudos.intersection(family_pseudos)


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_parse_fail(tmpdir):
    """Test the `PseudoPotentialFamily.create_from_folder` class method for file that fails to parse.

    Since the base pseudo potential class cannot really fail to parse, since there is no parsing, this would be
    difficult to test, however, the constructor parses the filename for the element, and that can fail if the filename
    has the incorrect format.
    """
    with open(os.path.join(str(tmpdir), 'Arr.upf'), 'wb'):
        pass

    with pytest.raises(exceptions.ParsingError, match=r'`.*` constructor did not define the element .*'):
        PseudoPotentialFamily.create_from_folder(str(tmpdir), 'label')


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_empty(tmpdir):
    """Test the `PseudoPotentialFamily.create_from_folder` class method for empty folder."""
    with pytest.raises(ValueError, match=r'no pseudo potentials were parsed from.*'):
        PseudoPotentialFamily.create_from_folder(str(tmpdir), 'label')


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_duplicate_element(tmpdir, filepath_pseudos):
    """Test the `PseudoPotentialFamily.create_from_folder` class method for folder containing duplicate element."""
    distutils.dir_util.copy_tree(filepath_pseudos(), str(tmpdir))

    with open(os.path.join(str(tmpdir), 'Ar.UPF'), 'wb'):
        pass

    with pytest.raises(ValueError, match=r'directory `.*` contains pseudo potentials with duplicate elements'):
        PseudoPotentialFamily.create_from_folder(str(tmpdir), 'label')


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_duplicate(filepath_pseudos):
    """Test that `PseudoPotentialFamily.create_from_folder` raises for duplicate label."""
    label = 'label'
    PseudoPotentialFamily(label=label).store()

    with pytest.raises(ValueError, match=r'the PseudoPotentialFamily `.*` already exists'):
        PseudoPotentialFamily.create_from_folder(filepath_pseudos(), label)


def test_parse_pseudos_from_directory_non_file(tmpdir):
    """Test the `PseudoPotentialFamily.parse_pseudos_from_directory` class method for folder containing a non-file.

    Note that a subdirectory containing the pseudos is fine, but if we find a directory and any other object at the
    base path, it should raise.
    """
    os.makedirs(os.path.join(str(tmpdir), 'directory'))
    with open(os.path.join(str(tmpdir), 'Ar.upf'), 'wb'):
        pass

    with pytest.raises(ValueError, match=r'dirpath `.*` contains at least one entry that is not a file'):
        PseudoPotentialFamily.parse_pseudos_from_directory(str(tmpdir))


def test_parse_pseudos_from_directory_non_file_nested(tmpdir):
    """Test the `PseudoPotentialFamily.parse_pseudos_from_directory` class method for folder containing a non-file.

    Note that a subdirectory containing the pseudos is fine, but if we find a directory and any other object at the
    base path, it should raise.
    """
    os.makedirs(os.path.join(str(tmpdir), 'pseudos', 'directory'))
    with open(os.path.join(str(tmpdir), 'pseudos', 'Ar.upf'), 'wb'):
        pass

    with pytest.raises(ValueError, match=r'dirpath `.*` contains at least one entry that is not a file'):
        PseudoPotentialFamily.parse_pseudos_from_directory(str(tmpdir))


@pytest.mark.usefixtures('clear_db')
def test_add_nodes(get_pseudo_family, get_pseudo_potential_data):
    """Test that `PseudoPotentialFamily.add_nodes` method."""
    family = get_pseudo_family(elements=('Rn',))
    assert family.count() == 1

    pseudos = get_pseudo_potential_data('Ar').store()
    family.add_nodes(pseudos)
    assert family.count() == 2

    pseudos = (get_pseudo_potential_data('Ne').store(),)
    family.add_nodes(pseudos)
    assert family.count() == 3

    pseudos = (get_pseudo_potential_data('He').store(), get_pseudo_potential_data('Kr').store())
    family.add_nodes(pseudos)
    assert family.count() == 5

    # Test for an unstored family
    family = PseudoPotentialFamily(label='label')
    with pytest.raises(exceptions.ModificationNotAllowed):
        family.add_nodes(pseudos)


@pytest.fixture
def nodes_unstored(get_pseudo_potential_data, request):
    """Dynamic fixture returning instances of `PseudoPotentialData` either isolated or as a list."""
    if request.param == 'single':
        return get_pseudo_potential_data()

    if request.param == 'tuple':
        return (get_pseudo_potential_data(),)

    return [get_pseudo_potential_data(), get_pseudo_potential_data('Ne')]


@pytest.mark.usefixtures('clear_db')
@pytest.mark.parametrize('nodes_unstored', ['single', 'tuple', 'list'], indirect=True)
def test_add_nodes_unstored(get_pseudo_family, nodes_unstored):
    """Test that `PseudoPotentialFamily.add_nodes` fails if one or more nodes are unstored."""
    family = get_pseudo_family(elements=('He',))
    count = family.count()

    with pytest.raises(ValueError, match='At least one of the provided nodes is unstored, stopping...'):
        family.add_nodes(nodes_unstored)

    assert family.count() == count


@pytest.fixture
def nodes_incorrect_type(get_pseudo_potential_data, request):
    """Dynamic fixture returning instances of `UpfData` either isolated or as a list."""
    if request.param == 'single':
        return get_pseudo_potential_data(entry_point='upf').store()

    if request.param == 'tuple':
        return (get_pseudo_potential_data(entry_point='upf').store(),)

    return [get_pseudo_potential_data().store(), get_pseudo_potential_data(entry_point='upf').store()]


@pytest.mark.usefixtures('clear_db')
@pytest.mark.parametrize('nodes_incorrect_type', ['single', 'tuple', 'list'], indirect=True)
def test_add_nodes_incorrect_type(get_pseudo_family, nodes_incorrect_type):
    """Test that `PseudoPotentialFamily.add_nodes` fails if one or more nodes has the incorrect type.

    Even though `UpfData` is a subclass of `PseudoPotentialData` it should still be refused because `add_nodes` checks
    for exact equality of the expected type and does not accept subclasses.
    """
    family = get_pseudo_family()
    count = family.count()

    with pytest.raises(TypeError, match=r'only nodes of type `.*` can be added'):
        family.add_nodes(nodes_incorrect_type)

    assert family.count() == count


@pytest.mark.usefixtures('clear_db')
def test_add_nodes_duplicate_element(get_pseudo_family, get_pseudo_potential_data):
    """Test that `PseudoPotentialFamily.add_nodes` fails if a pseudo is added whose element already exists."""
    family = get_pseudo_family(elements=('Ar',))
    pseudo = get_pseudo_potential_data('Ar').store()

    with pytest.raises(ValueError, match='element `Ar` already present in this family'):
        family.add_nodes(pseudo)


@pytest.mark.usefixtures('clear_db')
def test_pseudos(get_pseudo_potential_data):
    """Test the `PseudoPotentialFamily.pseudos` property."""
    pseudos = {
        'Ar': get_pseudo_potential_data('Ar').store(),
        'He': get_pseudo_potential_data('He').store(),
    }
    family = PseudoPotentialFamily(label='label').store()
    family.add_nodes(list(pseudos.values()))
    assert family.pseudos == pseudos


@pytest.mark.usefixtures('clear_db')
def test_pseudos_mutate(get_pseudo_family, get_pseudo_potential_data):
    """Test that `PseudoPotentialFamily.pseudos` property does not act as a setter."""
    family = get_pseudo_family()

    with pytest.raises(AttributeError):
        family.pseudos = {'He': get_pseudo_potential_data('He')}


@pytest.mark.usefixtures('clear_db')
def test_elements(get_pseudo_family):
    """Test the `PseudoPotentialFamily.elements` property."""
    elements = ['Ar', 'He']
    family = get_pseudo_family(elements=elements)
    assert sorted(family.elements) == elements

    family = PseudoPotentialFamily(label='empty').store()
    assert family.elements == []


@pytest.mark.usefixtures('clear_db')
def test_get_pseudo(get_pseudo_family):
    """Test the `PseudoPotentialFamily.get_pseudo` method."""
    element = 'Ar'
    family = get_pseudo_family(elements=(element,))

    assert family.get_pseudo(element) == family.pseudos[element]

    with pytest.raises(ValueError, match=r'family `.*` does not contain pseudo for element `.*`'):
        family.get_pseudo('He')


@pytest.mark.usefixtures('clear_db')
def test_get_pseudos_raise(get_pseudo_family, generate_structure):
    """Test the `PseudoPotentialFamily.get_pseudos` method when it is supposed to raise."""
    elements = ('Ar', 'He', 'Ne')
    structure = generate_structure(elements)
    family = get_pseudo_family(elements=elements[:2])  # Create family with only subset of the elements

    with pytest.raises(ValueError, match='have to specify one of the keyword arguments `elements` and `structure`.'):
        family.get_pseudos()

    with pytest.raises(ValueError, match='cannot specify both keyword arguments `elements` and `structure`.'):
        family.get_pseudos(elements=elements, structure=structure)

    with pytest.raises(ValueError, match='elements should be a list or tuple of symbols.'):
        family.get_pseudos(elements={'He', 'Ar'})

    with pytest.raises(ValueError, match='structure should be a `StructureData` instance.'):
        family.get_pseudos(structure={'He', 'Ar'})

    with pytest.raises(ValueError, match=r'family `.*` does not contain pseudo for element `.*`'):
        family.get_pseudos(elements=elements)

    with pytest.raises(ValueError, match=r'family `.*` does not contain pseudo for element `.*`'):
        family.get_pseudos(structure=structure)


@pytest.mark.usefixtures('clear_db')
def test_get_pseudos_list(get_pseudo_family):
    """Test the `PseudoPotentialFamily.get_pseudos` method when passing a list of elements."""
    elements = ('Ar', 'He', 'Ne')
    family = get_pseudo_family(elements=elements)

    pseudos = family.get_pseudos(elements=elements)
    assert isinstance(pseudos, dict)
    for element in elements:
        assert isinstance(pseudos[element], PseudoPotentialData)


@pytest.mark.usefixtures('clear_db')
def test_get_pseudos_structure(get_pseudo_family, generate_structure):
    """Test the `PseudoPotentialFamily.get_pseudos` method when passing a ``StructureData`` instance."""
    elements = ('Ar', 'He', 'Ne')
    structure = generate_structure(elements)
    family = get_pseudo_family(elements=elements)

    pseudos = family.get_pseudos(structure=structure)
    assert isinstance(pseudos, dict)
    for element in elements:
        assert isinstance(pseudos[element], PseudoPotentialData)


@pytest.mark.usefixtures('clear_db')
def test_get_pseudos_structure_kinds(get_pseudo_family, generate_structure):
    """Test the `PseudoPotentialFamily.get_pseudos` for ``StructureData`` with kind names including digits."""
    elements = ('Ar1', 'Ar2')
    structure = generate_structure(elements)
    family = get_pseudo_family(elements=elements)

    pseudos = family.get_pseudos(structure=structure)
    assert isinstance(pseudos, dict)
    for element in elements:
        assert isinstance(pseudos[element], PseudoPotentialData)



# -*- coding: utf-8 -*-
"""Module for data plugin to represent a pseudo potential in Psp8 format."""
from typing import BinaryIO
from aiida.common.constants import elements

from .pseudo import PseudoPotentialData

__all__ = ('Psp8Data',)


def parse_element(stream: BinaryIO):
    """Parse the content of the Psp8 file to determine the element.

    :param stream: a filelike object with the binary content of the file.
    :return: the symbol of the element following the IUPAC naming standard.
    """
    lines = stream.read().decode('utf-8')

    # Split the line at each new line character \n
    lines_splt = lines.splitlines()

    # Split the second line on white space
    lines_splt_space = lines_splt[1].split()

    try:
        atomic_number = int(float(lines_splt_space[0]))
    except (IndexError, ValueError) as val_err:
        raise ValueError('failed to parse the atomic number.') from val_err

    try:
        symbol = elements[atomic_number]['symbol']
    except KeyError as key_err:
        raise ValueError('the atomic number {atomic_number} is not supported.') from key_err

    return symbol


class Psp8Data(PseudoPotentialData):
    """Data plugin to represent a pseudo potential in Psp8 (Abinit) format."""

    def set_file(self, stream: BinaryIO, filename: str = None, **kwargs):  # pylint: disable=arguments-differ
        """Set the file content.

        :param stream: a filelike object with the binary content of the file.
        :param filename: optional explicit filename to give to the file stored in the repository.
        :raises ValueError: if the element symbol is invalid.
        """
        stream.seek(0)
        self.element = parse_element(stream)
        stream.seek(0)
        super().set_file(stream, filename, **kwargs)



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument,pointless-statement
"""Tests for the `SsspFamily` class."""
import pytest

from aiida_pseudo.groups.family import SsspConfiguration, SsspFamily


def test_type_string(clear_db):
    """Verify the `_type_string` class attribute is correctly set to the corresponding entry point name."""
    assert SsspFamily._type_string == 'pseudo.family.sssp'  # pylint: disable=protected-access


def test_default_configuration():
    """Test the `SsspFamily.default_configuration` class attribute."""
    assert isinstance(SsspFamily.default_configuration, SsspConfiguration)


def test_valid_configurations():
    """Test the `SsspFamily.valid_configurations` class attribute."""
    valid_configurations = SsspFamily.valid_configurations
    assert isinstance(valid_configurations, tuple)

    for entry in valid_configurations:
        assert isinstance(entry, SsspConfiguration)


def test_get_valid_labels():
    """Test the `SsspFamily.get_valid_labels` class method."""
    valid_labels = SsspFamily.get_valid_labels()
    assert isinstance(valid_labels, tuple)

    for entry in valid_labels:
        assert isinstance(entry, str)


def test_format_configuration_label():
    """Test the `SsspFamily.format_configuration_label` class method."""
    configuration = SsspConfiguration(1.1, 'PBE', 'efficiency')
    assert SsspFamily.format_configuration_label(configuration) == 'SSSP/1.1/PBE/efficiency'


def test_constructor():
    """Test that the `SsspFamily` constructor validates the label."""
    with pytest.raises(ValueError, match=r'the label `.*` is not a valid SSSP configuration label'):
        SsspFamily()

    with pytest.raises(ValueError, match=r'the label `.*` is not a valid SSSP configuration label'):
        SsspFamily(label='SSSP_1.1_PBE_efficiency')

    label = SsspFamily.format_configuration_label(SsspFamily.default_configuration)
    family = SsspFamily(label=label)
    assert isinstance(family, SsspFamily)



# -*- coding: utf-8 -*-
"""Command to install a pseudo potential family."""
import json
import os
import shutil
import tempfile

import click

from aiida.cmdline.utils import decorators, echo
from aiida.cmdline.params import options as options_core
from aiida.cmdline.params import types

from .params import options
from .root import cmd_root

URL_SSSP_BASE = 'https://legacy-archive.materialscloud.org/file/2018.0001/v4/'


@cmd_root.group('install')
def cmd_install():
    """Install pseudo potential families."""


@cmd_install.command('family')
@click.argument('archive', type=types.FileOrUrl(mode='rb'))
@click.argument('label', type=click.STRING)
@options_core.DESCRIPTION(help='Description for the family.')
@options.ARCHIVE_FORMAT()
@options.FAMILY_TYPE()
@options.TRACEBACK()
@decorators.with_dbenv()
def cmd_install_family(archive, label, description, archive_format, family_type, traceback):  # pylint: disable=too-many-arguments
    """Install a standard pseudo potential family from an ARCHIVE on the local file system or from a URL.

    The command will attempt to infer the archive format from the filename extension of the ARCHIVE. If this fails, the
    archive format can be specified explicitly with the archive format option, which will also display which formats
    are supported.

    By default, the command will create a base `PseudoPotentialFamily`, but the type can be changed with the family
    type option. If the base type is used, the pseudo potential files in the archive *have* to have filenames that
    strictly follow the format `ELEMENT.EXTENSION`, because otherwise the element cannot be determined automatically.
    """
    from .utils import attempt, create_family_from_archive

    # The `archive` is now either a `http.client.HTTPResponse` or a normal filelike object, so we get the original file
    # name in a different way.
    try:
        suffix = os.path.basename(archive.url)
    except AttributeError:
        suffix = os.path.basename(archive.name)

    # Write the content of the archive to a temporary file, because `create_family_from_archive` does currently not
    # accept filelike objects because the underlying `shutil.unpack_archive` does not. Likewise, `unpack_archive` will
    # attempt to deduce the archive format from the filename extension, so it is important we maintain the original
    # filename. Of course if this fails, users can specify the archive format explicitly wiht the corresponding option.
    with tempfile.NamedTemporaryFile(mode='w+b', suffix=suffix) as handle:
        shutil.copyfileobj(archive, handle)
        handle.flush()

        with attempt('unpacking archive and parsing pseudos... ', include_traceback=traceback):
            family = create_family_from_archive(family_type, label, handle.name, fmt=archive_format)

    family.description = description
    echo.echo_success(f'installed `{label}` containing {family.count()} pseudo potentials')


@cmd_install.command('sssp')
@options.VERSION(type=click.Choice(['1.0', '1.1']), default='1.1')
@options.FUNCTIONAL(type=click.Choice(['PBE', 'PBEsol']), default='PBE')
@options.PROTOCOL(type=click.Choice(['efficiency', 'precision']), default='efficiency')
@options.TRACEBACK()
@decorators.with_dbenv()
def cmd_install_sssp(version, functional, protocol, traceback):
    """Install an SSSP configuration.

    The SSSP configuration will be automatically downloaded from the Materials Cloud Archive entry to create a new
    `SsspFamily`.
    """
    # pylint: disable=too-many-locals
    import requests

    from aiida.common.files import md5_file
    from aiida.orm import Group, QueryBuilder

    from aiida_pseudo import __version__
    from aiida_pseudo.groups.family import SsspConfiguration, SsspFamily
    from .utils import attempt, create_family_from_archive

    configuration = SsspConfiguration(version, functional, protocol)
    label = SsspFamily.format_configuration_label(configuration)
    description = f'SSSP v{version} {functional} {protocol} installed with aiida-pseudo v{__version__}'

    if configuration not in SsspFamily.valid_configurations:
        echo.echo_critical(f'{version} {functional} {protocol} is not a valid SSSP configuration')

    if QueryBuilder().append(SsspFamily, filters={'label': label}).first():
        echo.echo_critical(f'{SsspFamily.__name__}<{label}> is already installed')

    with tempfile.TemporaryDirectory() as dirpath:

        url_archive = f'{URL_SSSP_BASE}/SSSP_{version}_{functional}_{protocol}.tar.gz'
        filepath_archive = os.path.join(dirpath, 'archive.tar.gz')

        url_metadata = f'{URL_SSSP_BASE}/SSSP_{version}_{functional}_{protocol}.json'
        filepath_metadata = os.path.join(dirpath, 'metadata.json')

        with attempt('downloading selected pseudo potentials archive... ', include_traceback=traceback):
            response = requests.get(url_archive)
            response.raise_for_status()
            with open(filepath_archive, 'wb') as handle:
                handle.write(response.content)
                handle.flush()
                description += f'\nArchive pseudos md5: {md5_file(filepath_archive)}'

        with attempt('downloading selected pseudo potentials metadata... ', include_traceback=traceback):
            response = requests.get(url_metadata)
            response.raise_for_status()
            with open(filepath_metadata, 'a+b') as handle:
                handle.write(response.content)
                handle.flush()
                handle.seek(0)
                metadata = json.load(handle)
                description += f'\nPseudo metadata md5: {md5_file(filepath_metadata)}'

        with attempt('unpacking archive and parsing pseudos... ', include_traceback=traceback):
            family = create_family_from_archive(SsspFamily, label, filepath_archive)

        cutoffs = {}

        for element, values in metadata.items():
            if family.get_pseudo(element).md5 != values['md5']:
                Group.objects.delete(family.pk)
                msg = f"md5 of pseudo for element {element} does not match that of the metadata {values['md5']}"
                echo.echo_critical(msg)

            cutoffs[element] = {'cutoff_wfc': values['cutoff_wfc'], 'cutoff_rho': values['cutoff_rho']}

        family.description = description
        family.set_cutoffs({'normal': cutoffs})

        echo.echo_success(f'installed `{label}` containing {family.count()} pseudo potentials')



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument,pointless-statement
"""Tests for the `Psp8Family` class."""
import pytest

from aiida_pseudo.groups.family.psp8 import Psp8Family


def test_type_string(clear_db):
    """Verify the `_type_string` class attribute is correctly set to the corresponding entry point name."""
    assert Psp8Family._type_string == 'pseudo.family.psp8'  # pylint: disable=protected-access


def test_construct(clear_db):
    """Test the construction of `Psp8Family` works."""
    family = Psp8Family(label='family').store()
    assert isinstance(family, Psp8Family)

    description = 'PseudoDojo description'
    family = Psp8Family(label='PseudoDojo/v0.4', description=description).store()
    assert isinstance(family, Psp8Family)
    assert family.description == description


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder(filepath_pseudos):
    """Test the `Psp8Family.create_from_folder` class method."""
    label = 'label'
    family = Psp8Family.create_from_folder(filepath_pseudos('psp8'), label)
    assert isinstance(family, Psp8Family)


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_duplicate(filepath_pseudos):
    """Test that `Psp8Family.create_from_folder` raises for duplicate label."""
    label = 'label'
    Psp8Family(label=label).store()

    with pytest.raises(ValueError, match=r'the Psp8Family `.*` already exists'):
        Psp8Family.create_from_folder(filepath_pseudos('psp8'), label)



# -*- coding: utf-8 -*-
# pylint: disable=undefined-variable
"""Module with group plugins to represent pseudo potential families."""
from .pseudo import *
from .psf import *
from .psml import *
from .psp8 import *
from .sssp import *
from .upf import *

__all__ = (pseudo.__all__ + psf.__all__ + psml.__all__ + psp8.__all__ + sssp.__all__ + upf.__all__)



# -*- coding: utf-8 -*-
"""Group to represent a pseudo potential family with pseudos in Psp8 format."""
from aiida.plugins import DataFactory

from .pseudo import PseudoPotentialFamily

__all__ = ('Psp8Family',)

Psp8Data = DataFactory('pseudo.psp8')  # pylint: disable=invalid-name


class Psp8Family(PseudoPotentialFamily):
    """Group to represent a pseudo potential family with pseudos in Psp8 format."""

    _pseudo_type = Psp8Data



# -*- coding: utf-8 -*-
"""Subclass of `Group` that serves as a base class for representing pseudo potential families."""
import os
import re
from typing import Union, List, Tuple, Mapping

from aiida.common import exceptions
from aiida.common.lang import classproperty, type_check
from aiida.orm import Group, QueryBuilder
from aiida.plugins import DataFactory

from aiida_pseudo.data.pseudo import PseudoPotentialData

__all__ = ('PseudoPotentialFamily',)

StructureData = DataFactory('structure')


class PseudoPotentialFamily(Group):
    """Group to represent a pseudo potential family.

    This is a base class that provides most of the functionality but does not actually define what type of pseudo
    potentials can be contained. Subclasses should define the `_pseudo_type` class attribute to the data type of the
    pseudo potentials that are accepted. This *has* to be a subclass of `PseudoPotentialData`.
    """

    _pseudo_type = PseudoPotentialData
    _pseudos = None

    def __repr__(self):
        """Represent the instance for debugging purposes."""
        return f'{self.__class__.__name__}<{self.pk or self.uuid}>'

    def __str__(self):
        """Represent the instance for human-readable purposes."""
        return f'{self.__class__.__name__}<{self.label}>'

    def __init__(self, *args, **kwargs):
        """Validate that the `_pseudo_type` class attribute is a subclass of `PseudoPotentialData`."""
        if not issubclass(self._pseudo_type, PseudoPotentialData):
            class_name = self._pseudo_type.__class__.__name__
            raise RuntimeError(f'`{class_name}` is not a subclass of `PseudoPotentialData`.')

        super().__init__(*args, **kwargs)

    @classproperty
    def pseudo_type(cls):  # pylint: disable=no-self-argument
        """Return the pseudo potential type that this family accepts.

        :return: the subclass of ``PseudoPotentialData`` that this family hosts nodes of.
        """
        return cls._pseudo_type

    @classmethod
    def parse_pseudos_from_directory(cls, dirpath):
        """Parse the pseudo potential files in the given directory into a list of data nodes.

        .. note:: The directory pointed to by `dirpath` should only contain pseudo potential files. Optionally, it can
            contain just a single directory, that contains all the pseudo potential files. If any other files are stored
            in the basepath or the subdirectory, that cannot be successfully parsed as pseudo potential files the method
            will raise a `ValueError`.

        :param dirpath: absolute path to a directory containing pseudo potentials.
        :return: list of data nodes
        :raises ValueError: if `dirpath` is not a directory or contains anything other than files.
        :raises ValueError: if `dirpath` contains multiple pseudo potentials for the same element.
        :raises ParsingError: if the constructor of the pseudo type fails for one of the files in the `dirpath`.
        """
        from aiida.common.exceptions import ParsingError

        pseudos = []

        if not os.path.isdir(dirpath):
            raise ValueError(f'`{dirpath}` is not a directory')

        dirpath_contents = os.listdir(dirpath)

        if len(dirpath_contents) == 1 and os.path.isdir(os.path.join(dirpath, dirpath_contents[0])):
            dirpath = os.path.join(dirpath, dirpath_contents[0])

        for filename in os.listdir(dirpath):
            filepath = os.path.join(dirpath, filename)

            if not os.path.isfile(filepath):
                raise ValueError(f'dirpath `{dirpath}` contains at least one entry that is not a file')

            try:
                with open(filepath, 'rb') as handle:
                    pseudo = cls._pseudo_type(handle, filename=filename)
            except ParsingError as exception:
                raise ParsingError(f'failed to parse `{filepath}`: {exception}') from exception
            else:
                if pseudo.element is None:
                    match = re.search(r'^([A-Za-z]{1,2})\.\w+', filename)
                    if match is None:
                        raise ParsingError(
                            f'`{cls._pseudo_type}` constructor did not define the element and could not parse a valid '
                            'element symbol from the filename `{filename}` either. It should have the format '
                            '`ELEMENT.EXTENSION`'
                        )
                    pseudo.element = match.group(1)
                pseudos.append(pseudo)

        if not pseudos:
            raise ValueError(f'no pseudo potentials were parsed from `{dirpath}`')

        elements = set(pseudo.element for pseudo in pseudos)

        if len(pseudos) != len(elements):
            raise ValueError(f'directory `{dirpath}` contains pseudo potentials with duplicate elements')

        return pseudos

    @classmethod
    def create_from_folder(cls, dirpath, label, description='', deduplicate=True):
        """Create a new `PseudoPotentialFamily` from the pseudo potentials contained in a directory.

        :param dirpath: absolute path to the folder containing the UPF files.
        :param label: label to give to the `PseudoPotentialFamily`, should not already exist.
        :param description: description to give to the family.
        :param deduplicate: if True, will scan database for existing pseudo potentials of same type and with the same
            md5 checksum, and use that instead of the parsed one.
        :raises ValueError: if a `PseudoPotentialFamily` already exists with the given name.
        """
        type_check(description, str, allow_none=True)

        try:
            cls.objects.get(label=label)
        except exceptions.NotExistent:
            family = cls(label=label, description=description)
        else:
            raise ValueError(f'the {cls.__name__} `{label}` already exists')

        parsed_pseudos = cls.parse_pseudos_from_directory(dirpath)
        family_pseudos = []

        for pseudo in parsed_pseudos:
            if deduplicate:
                query = QueryBuilder()
                query.append(cls.pseudo_type, subclassing=False, filters={'attributes.md5': pseudo.md5})
                existing = query.first()
                if existing:
                    pseudo = existing[0]
            family_pseudos.append(pseudo)

        # Only store the `Group` and the pseudo nodes now, such that we don't have to worry about the clean up in the
        # case that an exception is raised during creating them.
        family.store()
        family.add_nodes([pseudo.store() for pseudo in family_pseudos])

        return family

    def add_nodes(self, nodes):
        """Add a node or a set of nodes to the family.

        .. note: Each family instance can only contain a single pseudo potential for each element.

        :param nodes: a single `Node` or a list of `Nodes` of type `PseudoPotentialFamily._pseudo_type`. Note that
            subclasses of `_pseudo_type` are not accepted, only instances of that very type.
        :raises ModificationNotAllowed: if the family is not stored.
        :raises TypeError: if nodes are not an instance or list of instance of `PseudoPotentialFamily._pseudo_type`.
        :raises ValueError: if any of the nodes are not stored or their elements already exist in this family.
        """
        if not self.is_stored:
            raise exceptions.ModificationNotAllowed('cannot add nodes to an unstored group')

        if not isinstance(nodes, (list, tuple)):
            nodes = [nodes]

        if any([type(node) is not self._pseudo_type for node in nodes]):  # pylint: disable=unidiomatic-typecheck
            raise TypeError(f'only nodes of type `{self._pseudo_type}` can be added: {nodes}')

        pseudos = {}

        # Check for duplicates before adding any pseudo to the internal cache
        for pseudo in nodes:
            if pseudo.element in self.elements:
                raise ValueError(f'element `{pseudo.element}` already present in this family')
            pseudos[pseudo.element] = pseudo

        self.pseudos.update(pseudos)

        super().add_nodes(nodes)

    @property
    def pseudos(self):
        """Return the dictionary of pseudo potentials of this family indexed on the element symbol.

        :return: dictionary of element symbol mapping pseudo potentials
        """
        if self._pseudos is None:
            self._pseudos = {pseudo.element: pseudo for pseudo in self.nodes}

        return self._pseudos

    @property
    def elements(self):
        """Return the list of elements for which this family defines a pseudo potential.

        :return: list of element symbols
        """
        return list(self.pseudos.keys())

    def get_pseudo(self, element):
        """Return the pseudo potential for the given element.

        :param element: the element for which to return the corresponding pseudo potential.
        :return: pseudo potential instance if it exists
        :raises ValueError: if the family does not contain a pseudo potential for the given element
        """
        try:
            pseudo = self.pseudos[element]
        except KeyError:
            builder = QueryBuilder()
            builder.append(self.__class__, filters={'id': self.pk}, tag='group')
            builder.append(self._pseudo_type, filters={'attributes.element': element}, with_group='group')

            try:
                pseudo = builder.one()[0]
            except exceptions.MultipleObjectsError as exception:
                raise RuntimeError(f'family `{self.label}` contains multiple pseudos for `{element}`') from exception
            except exceptions.NotExistent as exception:
                raise ValueError(
                    f'family `{self.label}` does not contain pseudo for element `{element}`'
                ) from exception
            else:
                self.pseudos[element] = pseudo

        return pseudo

    def get_pseudos(
        self,
        *,
        elements: Union[List[str], Tuple[str]] = None,
        structure: StructureData = None,
    ) -> Mapping[str, StructureData]:
        """Return the mapping of kind names on pseudo potential data nodes for the given list of elements or structure.

        :param elements: list of element symbols.
        :param structure: the ``StructureData`` node.
        :return: dictionary mapping the kind names of a structure on the corresponding pseudo potential data nodes.
        :raises ValueError: if the family does not contain a pseudo for any of the elements of the given structure.
        """
        if elements is not None and structure is not None:
            raise ValueError('cannot specify both keyword arguments `elements` and `structure`.')

        if elements is None and structure is None:
            raise ValueError('have to specify one of the keyword arguments `elements` and `structure`.')

        if elements is not None and not isinstance(elements, (list, tuple)) and not isinstance(elements, StructureData):
            raise ValueError('elements should be a list or tuple of symbols.')

        if structure is not None and not isinstance(structure, StructureData):
            raise ValueError('structure should be a `StructureData` instance.')

        if structure is not None:
            return {kind.name: self.get_pseudo(kind.symbol) for kind in structure.kinds}

        return {element: self.get_pseudo(element) for element in elements}



# -*- coding: utf-8 -*-
"""Reusable options for CLI commands."""
import shutil

import click

from aiida.cmdline.params.options import OverridableOption
from .types import PseudoPotentialFamilyTypeParam

__all__ = ('VERSION', 'FUNCTIONAL', 'PROTOCOL', 'TRACEBACK', 'FAMILY_TYPE', 'ARCHIVE_FORMAT')

VERSION = OverridableOption(
    '-v', '--version', type=click.STRING, required=False, help='Select the version of the SSSP configuration.'
)

FUNCTIONAL = OverridableOption(
    '-f', '--functional', type=click.STRING, required=False, help='Select the functional of the SSSP configuration.'
)

PROTOCOL = OverridableOption(
    '-p', '--protocol', type=click.STRING, required=False, help='Select the protocol of the SSSP configuration.'
)

TRACEBACK = OverridableOption(
    '-t', '--traceback', is_flag=True, help='Include the stacktrace if an exception is encountered.'
)

FAMILY_TYPE = OverridableOption(
    '-T',
    '--family-type',
    type=PseudoPotentialFamilyTypeParam(),
    default='pseudo.family',
    show_default=True,
    help='Choose the type of pseudo potential family to create.'
)

ARCHIVE_FORMAT = OverridableOption(
    '-F', '--archive-format', type=click.Choice([fmt[0] for fmt in shutil.get_archive_formats()])
)



# -*- coding: utf-8 -*-
# pylint: disable=wrong-import-position,wildcard-import
"""Module for the command line interface."""
import click_completion

# Activate the completion of parameter types provided by the click_completion package
click_completion.init()

from .root import cmd_root
from .install import cmd_install, cmd_install_family, cmd_install_sssp
from .list import cmd_list
from .show import cmd_show



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument
"""Tests for the command `aiida-pseudo show`."""
from aiida.orm import Group
from aiida_pseudo.cli import cmd_show
from aiida_pseudo.groups.family import PseudoPotentialFamily, UpfFamily


def test_show(clear_db, run_cli_command, get_pseudo_family):
    """Test the `aiida-pseudo show` command."""
    family = get_pseudo_family()
    result = run_cli_command(cmd_show, [family.label])
    for node in family.nodes:
        assert node.md5 in result.output
        assert node.element in result.output
        assert node.filename in result.output


def test_show_argument_type(clear_db, run_cli_command, get_pseudo_family):
    """Test that `aiida-pseudo show` only accepts instances of `PseudoPotentialFamily` or subclasses as argument."""
    pseudo_family = get_pseudo_family(label='pseudo-family', cls=PseudoPotentialFamily)
    upf_family = get_pseudo_family(label='upf-family', cls=UpfFamily)
    normal_group = Group('normal-group').store()

    run_cli_command(cmd_show, [pseudo_family.label])
    run_cli_command(cmd_show, [upf_family.label])
    run_cli_command(cmd_show, [normal_group.label], raises=SystemExit)


def test_show_raw(clear_db, run_cli_command, get_pseudo_family):
    """Test the `-r/--raw` option."""
    family = get_pseudo_family()

    for option in ['-r', '--raw']:
        result = run_cli_command(cmd_show, [option, family.label])
        assert len(result.output_lines) == len(family.nodes)



# -*- coding: utf-8 -*-
"""Tests for the :py:`~aiida_pseudo.data.pseudo.psp8` module."""
import os

import pytest

from aiida.common.exceptions import ModificationNotAllowed
from aiida_pseudo.data.pseudo import Psp8Data


@pytest.mark.usefixtures('clear_db')
def test_constructor(filepath_pseudos):
    """Test the constructor."""
    for filename in os.listdir(filepath_pseudos('psp8')):
        with open(os.path.join(filepath_pseudos('psp8'), filename), 'rb') as handle:
            pseudo = Psp8Data(handle, filename=filename)
            assert isinstance(pseudo, Psp8Data)
            assert not pseudo.is_stored
            assert pseudo.element == filename.split('.')[0]


@pytest.mark.usefixtures('clear_db')
def test_set_file(filepath_pseudos, get_pseudo_potential_data):
    """Test the `Psp8Data.set_file` method.

    This method allows to change the file, as long as the node has not been stored yet. We need to verify that all the
    information, such as attributes are commensurate when it is stored.
    """
    pseudo = get_pseudo_potential_data(element='Ar', entry_point='psp8')
    assert pseudo.element == 'Ar'

    with open(os.path.join(filepath_pseudos('psp8'), 'He.psp8'), 'rb') as handle:
        pseudo.set_file(handle)
        assert pseudo.element == 'He'

        pseudo.store()
        assert pseudo.element == 'He'

        with pytest.raises(ModificationNotAllowed):
            pseudo.set_file(handle)



# -*- coding: utf-8 -*-
"""Module for data plugin to represent a pseudo potential in PSF format."""
import re
from typing import BinaryIO

from .pseudo import PseudoPotentialData

__all__ = ('PsfData',)

REGEX_ELEMENT = re.compile(r"""\s*(?P<element>[a-zA-Z]{1}[a-z]?)\s+.*""")


def parse_element(stream: BinaryIO):
    """Parse the content of the PSF file to determine the element.

    :param stream: a filelike object with the binary content of the file.
    :return: the symbol of the element following the IUPAC naming standard.
    """
    lines = stream.read().decode('utf-8')
    line = lines.split('\n')[0]
    match = REGEX_ELEMENT.search(line)

    if match:
        return match.group('element').capitalize()

    raise ValueError('could not parse the element from the PSF content.')


class PsfData(PseudoPotentialData):
    """Data plugin to represent a pseudo potential in PSF format."""

    def set_file(self, stream: BinaryIO, filename: str = None, **kwargs):  # pylint: disable=arguments-differ
        """Set the file content.

        :param stream: a filelike object with the binary content of the file.
        :param filename: optional explicit filename to give to the file stored in the repository.
        :raises ValueError: if the element symbol is invalid.
        """
        stream.seek(0)
        self.element = parse_element(stream)
        stream.seek(0)
        super().set_file(stream, filename, **kwargs)



# -*- coding: utf-8 -*-
# pylint: disable=undefined-variable
"""Module for command line interface parameters."""
from .options import *
from .types import *

__all__ = (options.__all__ + types.__all__)



# -*- coding: utf-8 -*-
"""Module for data plugin to represent a pseudo potential in UPF format."""
import re
from typing import BinaryIO, Union

from .pseudo import PseudoPotentialData

__all__ = ('UpfData',)

REGEX_ELEMENT_V1 = re.compile(r"""(?P<element>[a-zA-Z]{1,2})\s+Element""")
REGEX_ELEMENT_V2 = re.compile(r"""\s*element\s*=\s*['"]\s*(?P<element>[a-zA-Z]{1,2})\s*['"].*""")

PATTERN_FLOAT = r'[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?'
REGEX_Z_VALENCE_V1 = re.compile(r"""(?P<z_valence>""" + PATTERN_FLOAT + r""")\s+Z valence""")
REGEX_Z_VALENCE_V2 = re.compile(r"""\s*z_valence\s*=\s*['"]\s*(?P<z_valence>""" + PATTERN_FLOAT + r""")\s*['"].*""")


def parse_element(content: str):
    """Parse the content of the UPF file to determine the element.

    :param stream: a filelike object with the binary content of the file.
    :return: the symbol of the element following the IUPAC naming standard.
    """
    for regex in [REGEX_ELEMENT_V2, REGEX_ELEMENT_V1]:

        match = regex.search(content)

        if match:
            return match.group('element')

    raise ValueError(f'could not parse the element from the UPF content: {content}')


def parse_z_valence(content: str) -> int:
    """Parse the content of the UPF file to determine the Z valence.

    :param stream: a filelike object with the binary content of the file.
    :return: the Z valence.
    """
    for regex in [REGEX_Z_VALENCE_V2, REGEX_Z_VALENCE_V1]:

        match = regex.search(content)

        if match:
            z_valence = match.group('z_valence')

            try:
                z_valence = float(z_valence)
            except ValueError as exception:
                raise ValueError(f'parsed value for the Z valence `{z_valence}` is not a valid number.') from exception

            if int(z_valence) != z_valence:
                raise ValueError(f'parsed value for the Z valence `{z_valence}` is not an integer.')

            return int(z_valence)

    raise ValueError(f'could not parse the Z valence from the UPF content: {content}')


class UpfData(PseudoPotentialData):
    """Data plugin to represent a pseudo potential in UPF format."""

    _key_z_valence = 'z_valence'

    def set_file(self, stream: BinaryIO, filename: str = None, **kwargs):  # pylint: disable=arguments-differ
        """Set the file content.

        :param stream: a filelike object with the binary content of the file.
        :param filename: optional explicit filename to give to the file stored in the repository.
        :raises ValueError: if the element symbol is invalid.
        """
        stream.seek(0)

        content = stream.read().decode('utf-8')
        self.element = parse_element(content)
        self.z_valence = parse_z_valence(content)

        stream.seek(0)
        super().set_file(stream, filename, **kwargs)

    @property
    def z_valence(self) -> Union[int, None]:
        """Return the Z valence.

        :return: the Z valence.
        """
        return self.get_attribute(self._key_z_valence, None)

    @z_valence.setter
    def z_valence(self, value: int):
        """Set the Z valence.

        :param value: the Z valence.
        :raises ValueError: if the value is not a positive integer.
        """
        if not isinstance(value, int) or value < 0:
            raise ValueError(f'`{value}` is not a positive integer')

        self.set_attribute(self._key_z_valence, value)



# -*- coding: utf-8 -*-
"""Subclass of `UpfFamily` designed to represent an SSSP configuration."""
from collections import namedtuple
from typing import Sequence

from .upf import UpfFamily

__all__ = ('SsspConfiguration', 'SsspFamily')

SsspConfiguration = namedtuple('SsspConfiguration', ['version', 'functional', 'protocol'])


class SsspFamily(UpfFamily):
    """Subclass of `UpfFamily` designed to represent an SSSP configuration.

    The `SsspFamily` is essentially a `UpfFamily` with some additional constraints. It can only be used to contain the
    pseudo potentials and corresponding metadata of an official SSSP configuration.
    """

    label_template = 'SSSP/{version}/{functional}/{protocol}'
    default_configuration = SsspConfiguration('1.1', 'PBE', 'efficiency')
    valid_configurations = (
        SsspConfiguration('1.0', 'PBE', 'efficiency'),
        SsspConfiguration('1.0', 'PBE', 'precision'),
        SsspConfiguration('1.1', 'PBE', 'efficiency'),
        SsspConfiguration('1.1', 'PBE', 'precision'),
        SsspConfiguration('1.1', 'PBEsol', 'efficiency'),
        SsspConfiguration('1.1', 'PBEsol', 'precision'),
    )

    @classmethod
    def get_valid_labels(cls) -> Sequence[str]:
        """Return the tuple of labels of all valid SSSP configurations."""
        return tuple(cls.format_configuration_label(configuration) for configuration in cls.valid_configurations)

    @classmethod
    def format_configuration_label(cls, configuration: SsspConfiguration) -> str:
        """Format a label for an `SsspFamily` with the required syntax.

        :param configuration: the SSSP configuration
        :return: label
        """
        return cls.label_template.format(
            version=configuration.version, functional=configuration.functional, protocol=configuration.protocol
        )

    def __init__(self, label=None, **kwargs):
        """Construct a new instance, validating that the label matches the required format."""
        if label not in self.get_valid_labels():
            raise ValueError(f'the label `{label}` is not a valid SSSP configuration label.')

        super().__init__(label=label, **kwargs)



# -*- coding: utf-8 -*-
# pylint: disable=unused-argument,pointless-statement
"""Tests for the `PsfFamily` class."""
import pytest

from aiida_pseudo.groups.family.psf import PsfFamily


def test_type_string(clear_db):
    """Verify the `_type_string` class attribute is correctly set to the corresponding entry point name."""
    assert PsfFamily._type_string == 'pseudo.family.psf'  # pylint: disable=protected-access


def test_construct(clear_db):
    """Test the construction of `PsfFamily` works."""
    family = PsfFamily(label='family').store()
    assert isinstance(family, PsfFamily)

    description = 'SSSP description'
    family = PsfFamily(label='SSSP/v1.1', description=description).store()
    assert isinstance(family, PsfFamily)
    assert family.description == description


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder(filepath_pseudos):
    """Test the `PsfFamily.create_from_folder` class method."""
    label = 'label'
    family = PsfFamily.create_from_folder(filepath_pseudos('psf'), label)
    assert isinstance(family, PsfFamily)


@pytest.mark.usefixtures('clear_db')
def test_create_from_folder_duplicate(filepath_pseudos):
    """Test that `PsfFamily.create_from_folder` raises for duplicate label."""
    label = 'label'
    PsfFamily(label=label).store()

    with pytest.raises(ValueError, match=r'the PsfFamily `.*` already exists'):
        PsfFamily.create_from_folder(filepath_pseudos('psf'), label)



# -*- coding: utf-8 -*-
# pylint: disable=no-self-use
"""Custom parameter types for command line interface commands."""
import click

from aiida.cmdline.params.types import GroupParamType

__all__ = ('PseudoPotentialFamilyTypeParam', 'PseudoPotentialFamilyParam')


class PseudoPotentialFamilyParam(GroupParamType):
    """Parameter type for `click` commands to define an instance of a `PseudoPotentialFamily`."""

    name = 'pseudo_family'


class PseudoPotentialFamilyTypeParam(click.ParamType):
    """Parameter type for `click` commands to define a subclass of `PseudoPotentialFamily`."""

    name = 'pseudo_family_type'

    def convert(self, value, _, __):
        """Convert the entry point name to the corresponding class.

        :param value: entry point name that should correspond to subclass of `PseudoPotentialFamily` group plugin
        :return: the `PseudoPotentialFamily` subclass
        :raises: `click.BadParameter` if the entry point cannot be loaded or is not subclass of `PseudoPotentialFamily`
        """
        from aiida.common import exceptions
        from aiida.plugins import GroupFactory
        from aiida_pseudo.groups.family import PseudoPotentialFamily

        try:
            family_type = GroupFactory(value)
        except exceptions.EntryPointError as exception:
            raise click.BadParameter(f'`{value}` is not an existing group plugin.') from exception

        if not issubclass(family_type, PseudoPotentialFamily):
            raise click.BadParameter(f'`{value}` entry point is not a subclass of `PseudoPotentialFamily`.')

        PseudoPotentialFamily.entry_point = value

        return family_type

    def complete(self, _, incomplete):
        """Return possible completions based on an incomplete value.

        :returns: list of tuples of valid entry points (matching incomplete) and a description
        """
        from aiida.plugins.entry_point import get_entry_point_names
        entry_points = get_entry_point_names('aiida.groups')
        return [(ep, '') for ep in entry_points if (ep.startswith('pseudo.family') and ep.startswith(incomplete))]



# -*- coding: utf-8 -*-
"""Tests for `aiida-pseudo install`."""
import pytest

from aiida.orm import QueryBuilder

from aiida_pseudo.cli import cmd_install_family, cmd_install_sssp
from aiida_pseudo.groups.family import PseudoPotentialFamily, UpfFamily


@pytest.mark.usefixtures('clear_db')
def test_install_family(run_cli_command, get_pseudo_archive):
    """Test `aiida-pseudo install family`."""
    label = 'family'
    description = 'description'
    filepath_archive = next(get_pseudo_archive())
    options = ['-D', description, filepath_archive, label]

    result = run_cli_command(cmd_install_family, options)
    assert f'installed `{label}`' in result.output
    assert PseudoPotentialFamily.objects.count() == 1

    family = PseudoPotentialFamily.objects.get(label=label)
    assert family.__class__ is PseudoPotentialFamily
    assert family.description == description
    assert len(family.pseudos) != 0


@pytest.mark.usefixtures('clear_db')
def test_install_family_url(run_cli_command):
    """Test `aiida-pseudo install family` when installing from a URL.

    When a URL is passed, the parameter converts it into a `http.client.HTTPResponse`, which is not trivial to mock so
    instead we use an actual URL, which is slow, but is anyway already tested indirectly in `test_install_sssp`.
    """
    label = 'family'
    description = 'description'
    filepath_archive = 'https://legacy-archive.materialscloud.org/file/2018.0001/v4/SSSP_1.0_PBE_efficiency.tar.gz'
    options = ['-D', description, filepath_archive, label, '-T', 'pseudo.family.upf']

    result = run_cli_command(cmd_install_family, options)
    assert f'installed `{label}`' in result.output
    assert UpfFamily.objects.count() == 1

    family = UpfFamily.objects.get(label=label)
    assert family.__class__ is UpfFamily
    assert family.description == description
    assert len(family.pseudos) != 0


@pytest.mark.usefixtures('clear_db')
def test_install_family_upf(run_cli_command, get_pseudo_archive):
    """Test `aiida-pseudo install family` to install a `UpfFamily`."""
    label = 'family'
    description = 'description'
    filepath_archive = next(get_pseudo_archive())
    options = ['-D', description, '-T', 'pseudo.family.upf', filepath_archive, label]

    result = run_cli_command(cmd_install_family, options)
    assert f'installed `{label}`' in result.output
    assert UpfFamily.objects.count() == 1

    family = UpfFamily.objects.get(label=label)
    assert family.__class__ is UpfFamily
    assert family.description == description
    assert len(family.pseudos) != 0


@pytest.mark.usefixtures('clear_db')
def test_install_sssp(run_cli_command):
    """Test the `aiida-pseudo install sssp` command."""
    from aiida_pseudo import __version__
    from aiida_pseudo.groups.family import SsspFamily

    result = run_cli_command(cmd_install_sssp)
    assert 'installed `SSSP/' in result.output
    assert QueryBuilder().append(SsspFamily).count() == 1

    family = QueryBuilder().append(SsspFamily).one()[0]
    assert family.get_cutoffs is not None
    assert f'SSSP v1.1 PBE efficiency installed with aiida-pseudo v{__version__}' in family.description
    assert 'Archive pseudos md5: 4803ce9fd1d84c777f87173cd4a2de33' in family.description
    assert 'Pseudo metadata md5: 0d5d6c2c840383c7c4fc3a99b5dc3001' in family.description

    result = run_cli_command(cmd_install_sssp, raises=SystemExit)
    assert 'is already installed' in result.output



# -*- coding: utf-8 -*-
"""Command line interface `aiida-pseudo`."""
import click

from aiida.cmdline.params import options, types


@click.group('aiida-pseudo', context_settings={'help_option_names': ['-h', '--help']})
@options.PROFILE(type=types.ProfileParamType(load_profile=True))
def cmd_root(profile):  # pylint: disable=unused-argument
    """CLI for the `aiida-pseudo` plugin."""
